def config_pandas_display():
    """
    Configure display
    """
    import pandas as pd

    global pd
    pd.set_option("display.max_columns", 500)
    # pd.set_option("display.max_colwidth", -1)
    pd.set_option("display.max_colwidth", 500)
    # pd.set_option("display.max_rows", 200)
    pd.set_option("display.expand_frame_repr", True)


def add_type_columns(dataframe):
    """
    Returns dataframe where each columns has their dtype column right next to it

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Dataframe to add type columns

    Examples
    --------
    >>> values = [[0, '0', [], {}, None, ""]]
    >>> cols = ["int_", "str_", "list_", "set_", "none_", "_"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> df
       int_ str_ list_ set_ none_ _
    0     0    0    []   {}  None
    >>> add_type_columns(df)
       int_      int__type str_      str__type list_      list__type set_       set__type none_          none__type _         __type
    0     0  <class 'int'>    0  <class 'str'>    []  <class 'list'>   {}  <class 'dict'>  None  <class 'NoneType'>    <class 'str'>

    Returns
    -------
    df : pandas.DataFrame
    """
    import pandas as pd

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument must be a dataframe ...")

    df = dataframe.copy()
    suffix = "_type"
    for column in df.columns:
        # To prevent creating (1) duplicated column (2) column with ever-increasing suffix [xxx_type, xxx_type_type ...]
        new_col = column + suffix
        if not (new_col in df.columns or column.endswith(suffix)):
            new_col_loc = df.columns.get_loc(column) + 1
            new_col_value = df[column].apply(lambda x: type(x))
            df.insert(new_col_loc, column=new_col, value=new_col_value)
            # df[new_col] = df[column].apply(lambda x: type(x))
    return df


def chunking_dataframe(dataframe, chunk_size):
    """
    Return list of dataframes split from given dataframe and chunk size

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Original dataframe
    chunk_size : int
        The chunk size of each smaller dataframes

    Examples
    --------
    >>> chunking_dataframe(pd.DataFrame(), 2)
    [Empty DataFrame
    Columns: []
    Index: []]

    >>> values = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    >>> cols = ["test"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> chunking_dataframe(df, 5)
    [   test
    0     1
    1     2
    2     3
    3     4
    4     5,    test
    5     6
    6     7
    7     8
    8     9
    9    10, Empty DataFrame
    Columns: [test]
    Index: []]

    >>> chunking_dataframe(df, 4)
    [   test
    0     1
    1     2
    2     3
    3     4,    test
    4     5
    5     6
    6     7
    7     8,    test
    8     9
    9    10]

    Returns
    -------
    list_ : list
        [dataframe, dataframe2, ...]
    """
    import numpy as np
    import pandas as pd

    list_ = None
    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument must be a dataframe ...")
    if not isinstance(chunk_size, int):
        raise TypeError("Argument must be a int ...")

    indices = index_marks(dataframe.shape[0], chunk_size)
    list_ = np.split(dataframe, indices)
    return list_


def clean_none(dataframe, clean_variation=True):
    """
    Return a dataframe from given dataframe with standardized None

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Dataframe

    Examples
    --------
    >>> from base import get_none_variation
    >>> df = pd.DataFrame({"c1": get_none_variation()})
    >>> df
          c1
    0   None
    1   none
    2   None
    3   NONE
    4   null
    5   Null
    6   NULL
    7     na
    8     Na
    9     nA
    10    NA
    11   N.A
    12  N.A.
    13   nil
    14   Nil
    15   NIL
    >>> clean_none(df)
          c1
    0   None
    1   None
    2   None
    3   None
    4   None
    5   None
    6   None
    7   None
    8   None
    9   None
    10  None
    11  None
    12  None
    13  None
    14  None
    15  None
    >>> df = pd.DataFrame({"c1": [""]})
    >>> clean_none(df)
         c1
    0  None
    >>> df = pd.DataFrame({"c1": ["Nathing"]})
    >>> clean_none(df)
            c1
    0  Nathing
    >>> df = pd.DataFrame({"c1": ["Na "]})
    >>> clean_none(df)
        c1
    0  Na

    Returns
    -------
    df : pandas.DataFrame
    """
    import numpy as np
    import pandas as pd
    from data_manipulation.base import get_none_variation

    df = dataframe.copy()
    df = df.replace(r"^\s*$", np.nan, regex=True)
    if clean_variation:
        non_variations = get_none_variation()
        df = df.replace(non_variations, np.nan)
    df = df.where(pd.notnull(df), None)
    return df


def compare_all_list_items(list_):
    """
    Return dataframe of given list's Cartesian product

    Parameters
    ----------
    list_ : list
        A Python list

    Examples
    --------
    >>> compare_all_list_items([])
    Empty DataFrame
    Columns: [item1, item2, item1_eq_item2]
    Index: []

    >>> compare_all_list_items([1, 1, 2, 3])
      item1 item2  item1_eq_item2
    0     1     1            True
    1     1     2           False
    2     1     3           False
    3     1     2           False
    4     1     3           False
    5     2     3           False

    Returns
    -------
    df : pandas.DataFrame
    """
    import pandas as pd
    from itertools import combinations

    if not isinstance(list_, list):
        raise TypeError("Argument must be a list ...")

    df = []
    for a, b in combinations(list_, 2):
        row = {"item1": repr(a), "item2": repr(b), "item1_eq_item2": a == b}
        df.append(row)

    df = pd.DataFrame(df, columns=["item1", "item2", "item1_eq_item2"])
    return df


def compare_dataframes(dataframe, dataframe2):
    """
    Print the difference between the two given dataframes

    Parameters
    ----------
    dataframe : pandas.DataFrame
        1st dataframe
    dataframe2 : pandas.DataFrame
        2nd dataframe

    Examples
    --------
    >>> df1 = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    >>> df1
       a  b
    0  1  3
    1  2  4
    >>> df2 = df1.copy()
    >>> df2["a"] = df2["a"].apply(lambda x: x*2)
    >>> df2
       a  b
    0  2  3
    1  4  4
    >>> compare_dataframes(df1, df2)
    ================
    Dataframe length
    ================
    df1 length: 2
    df2 length: 2
    same df length: True
    <BLANKLINE>
    ----------------
    COMPARE COLUMNS
    ----------------
    df1 non-null a: 2
    df2 non-null a: 2
    df1 & df2 has the same length
    df1 non-null b: 2
    df2 non-null b: 2
    df1 & df2 has the same length

    Returns
    -------
    None
    """
    import pandas as pd

    if not all(isinstance(x, pd.DataFrame) for x in [dataframe, dataframe2]):
        raise TypeError("Arguments 1 & 2 must be pandas dataframes ...")

    print("=" * len("Dataframe length"))
    print("Dataframe length")
    print("=" * len("Dataframe length"))
    print(f"df1 length: {len(dataframe)}")
    print(f"df2 length: {len(dataframe)}")
    print(f"same df length: {len(dataframe) == len(dataframe2)}\n")
    print("----------------")
    print("COMPARE COLUMNS")
    print("----------------")
    for column in dataframe.columns:
        pd_df = dataframe[dataframe[column].notnull()]
        pd_df2 = None
        print(f"df1 non-null {column}: {len(pd_df)}")

        try:
            pd_df2 = dataframe2[dataframe2[column].notnull()]
        except KeyError:
            print(f"df2 {column}: does not exist")

        if pd_df2 is not None:
            print(f"df2 non-null {column}: {len(pd_df2)}")

            if len(pd_df) == len(pd_df2):
                print("df1 & df2 has the same length")
            elif len(pd_df2) > len(pd_df):
                print("df2 > df1, check")
            else:
                strings = ["null", "NA"]
                null_count = len(dataframe[dataframe[column].isnull()])
                string_count = len(dataframe[dataframe[column].isin(strings)])
                empty_string_count = len(dataframe[dataframe[column].str.len() == 0])
                non_null_count = len(dataframe) - null_count - string_count - empty_string_count
                print(f"""
                df1
                - null ({null_count})
                - string null ({string_count})
                - empty string ({empty_string_count})
                = {non_null_count}
                == df2 ({len(pd_df2)}): {non_null_count == len(pd_df2)}
                """)


def dtypes_dictionary(dataframe):
    """
    Return dictionary of dtypes from given dataframe's columns

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Base dataframe

    Examples
    --------
    >>> dtypes_dictionary(pd.DataFrame())
    Traceback (most recent call last):
      ...
    ValueError: Argument can't be a empty dataframe ...

    >>> values = [[0, '0', [], {}, None, ""]]
    >>> cols = ["int_", "str_", "list_", "set_", "none_", "_"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> df
       int_ str_ list_ set_ none_ _
    0     0    0    []   {}  None
    >>> dtypes_dictionary(df)
    {<class 'numpy.int64'>: ['int_'], <class 'str'>: ['str_', '_'], <class 'list'>: ['list_'], <class 'dict'>: ['set_'], <class 'NoneType'>: ['none_']}

    Returns
    -------
    dict_ : dict
        {str: [column1, column2, ...], int: [column5, ...], ...}
    """
    import pandas as pd

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument must be a dataframe ...")
    if len(dataframe) == 0:
        raise ValueError("Argument can't be a empty dataframe ...")

    dict_ = {}
    for index, value in dataframe.loc[0].iteritems():
        dtype = type(value)
        if dtype not in dict_:
            dict_[dtype] = [index]
        else:
            dict_[dtype].append(index)

    return dict_


def head_tail(dataframe, n=5):
    """
    Return dataframe from given dataframe with n head and tail

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Base dataframe
    n : int, optional
        Number of rows

    Examples
    --------
    >>> values = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    >>> cols = ["test"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> head_tail(df)
       test
    0     1
    1     2
    2     3
    3     4
    4     5
    5     6
    6     7
    7     8
    8     9
    9    10

    Returns
    -------
    df : pandas.DataFrame
    """
    import pandas as pd

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument must be a dataframe ...")

    df = dataframe.head(n).append(dataframe.tail(n))
    return df


def index_marks(n_rows, chunk_size):
    """
    Return Python range from given
        1. number of rows
        2. desired chunk size

    Parameters
    ----------
    n_rows : int
        The number of rows
    chunk_size : int
        The desire chunk size

    Examples
    --------
    >>> index_marks(10, 3)
    range(3, 12, 3)
    >>> list(index_marks(10, 3))
    [3, 6, 9]

    Returns
    -------
    range_ : range
    """
    if not all(isinstance(x, int) for x in [n_rows, chunk_size]):
        raise TypeError("Arguments 1 & 2 must be int ...")

    p1 = 1 * chunk_size
    p2 = (n_rows // chunk_size + 1) * chunk_size
    p3 = chunk_size
    range_ = range(p1, p2, p3)
    return range_


def print_dataframe_overview(dataframe, stats=False):
    """
    Print the given dataframe's columns' value_counts() in console

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Base dataframe
    stats : bool, optional
        Display statistics

    Examples
    --------
    >>> print_dataframe_overview(pd.DataFrame())

    >>> data = {"int_": [1, 1, 2], "str_": ["a", "b", "b"]}
    >>> df = pd.DataFrame(data)
    >>> print_dataframe_overview(df)
    ==============================
    int_
    ==============================
    Unique elements: 2
    Null elements: 0
       count    percent
    1      2  66.666667
    2      1  33.333333
    ==============================
    str_
    ==============================
    Unique elements: 2
    Null elements: 0
       count    percent
    b      2  66.666667
    a      1  33.333333

    Returns
    -------
    None
    """
    import pandas as pd
    from IPython.display import display

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument must be a dataframe ...")

    for column in dataframe.columns:
        print("=" * 30)
        print(column)
        print("=" * 30)
        try:
            print(f"Unique elements: {dataframe[column].nunique()}")
            print(f"Null elements: {len(dataframe[dataframe[column].isnull()])}")
            if stats:
                print(f"Minimum: {min(dataframe[column])}")
                print(f"Maximum: {max(dataframe[column])}")
            display(series_count(dataframe[column]).head(20))
        except:
            print(f"Unable to get value_counts of {column} ...\n")


def series_count(series):
    """
    Return dataframe with enhanced series.value_counts()

    Parameters
    ----------
    series : pd.Series
        The series to get value_counts()

    Examples
    --------
    >>> values = [1, 1, 1, 2, 2, 3]
    >>> cols = ["test"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> series_count(df["test"])
       count    percent
    1      3  50.000000
    2      2  33.333333
    3      1  16.666667

    Returns
    -------
    df : pandas.DataFrame
    """
    import pandas as pd

    if not isinstance(series, pd.Series):
        raise TypeError("Argument must be a series ...")

    df = series.value_counts().to_frame(name="count")
    total = df.sum()["count"]
    df["percent"] = df["count"] / total * 100
    return df


def series_to_columns(dataframe, column):
    """
    Return a dataframe converted from given dataframe's dtype column

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Dataframe
    column : pd.Series
        Series with dict dtype

    Examples
    --------
    >>> values = [[1, {'a':1, 'b':2}]]
    >>> cols = ["c1", "c2"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> df
       c1                c2
    0   1  {'a': 1, 'b': 2}
    >>> series_to_columns(df, "c2")
       c1  a  b
    0   1  1  2

    Returns
    -------
    df : pandas.DataFrame
    """
    import pandas as pd

    df = pd.concat([dataframe.drop([column], axis=1), dataframe[column].apply(pd.Series)], axis=1)
    return df


def split_dataframe(dataframe, uuid, columns):
    """
    Return tuple of dataframes from given dataframe, linked by uuid and have different columns

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Original dataframe
    uuid : str
        The uuid that can join the splited dataframes
    columns : list
        The list of columns to split from original dataframe

    Examples
    --------
    >>> df = pd.DataFrame({"a": [1, 2], "b": [2, 3], "uuid": ["abc123", "efg456"]})
    >>> df
       a  b    uuid
    0  1  2  abc123
    1  2  3  efg456
    >>> df1, df2 = split_dataframe(df, "uuid", ["b"])
    >>> df1
       a    uuid
    0  1  abc123
    1  2  efg456
    >>> df2
       b    uuid
    0  2  abc123
    1  3  efg456

    Returns
    -------
    df1: pandas.DataFrame
        Dataframe without specific columns
    df2: pandas.DataFrame
        Datafrane with specific columns and uuid

    (df1, df2)
    """
    import pandas as pd

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument 1 must be pandas.DataFrame ...")
    if not isinstance(uuid, str):
        raise TypeError("Argument 2 must be str ...")
    if not isinstance(columns, list):
        raise TypeError("Argument 3 must be list ...")

    df1 = dataframe[dataframe.columns[~dataframe.columns.isin(columns)]].copy()
    df2 = dataframe[columns + [uuid]].copy()
    return df1, df2


def split_left_merged_dataframe(dataframe, dataframe2, columns):
    """
    Return tuple of dataframes split by merged_type after left merging two given dataframes at columns

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Based dataframe
    dataframe2 : pandas.DataFrame
        Dataframe to merged with
    columns : list
        Columns to join on

    Examples
    --------
    >>> df1 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo', "zen"], 'value': [1, 2, 3, 5, 9]})
    >>> df2 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'], 'value': [5, 6, 7, 8]})
    >>> df_both, df_left = split_left_merged_dataframe(df1, df2, ["key"])
                count    percent
    both            6  85.714286
    left_only       1  14.285714
    right_only      0   0.000000
    >>> df_both
       key  value  value_y _merge
    0  foo      1      5.0   both
    1  foo      1      8.0   both
    2  bar      2      6.0   both
    3  baz      3      7.0   both
    4  foo      5      5.0   both
    5  foo      5      8.0   both
    >>> df_left
       key  value  value_y     _merge
    6  zen      9      NaN  left_only


    Returns
    -------
    df_both : pandas.DataFrame
        Dataframe with merged keys on both input dataframes
    df_left : pandas.DataFrame
        Dataframe whose merge keys only appear in left / base dataframe

    (df_both, df_left)
    """
    import pandas as pd
    from IPython.display import display

    if not all(isinstance(x, pd.DataFrame) for x in [dataframe, dataframe2]):
        raise TypeError("Arguments 1 & 2 must be pandas dataframes ...")
    if not isinstance(columns, list):
        raise TypeError("Argument must be a list ...")

    df_merged = dataframe.merge(dataframe2, on=columns, how="left", indicator=True, suffixes=("", "_y"))
    df_both = df_merged[df_merged["_merge"] == "both"].copy()
    df_left = df_merged[df_merged["_merge"] == "left_only"].copy()
    display(series_count(df_merged["_merge"]))
    return df_both, df_left


def to_excel_keep_url(filepath, dataframe):
    """
    Save dataframe as excel, without converting urls to strings.
        - Excel auto converts string to urls
        - Excel limit of 65,530 urls per worksheet
        - Excel does not keep url > 255 characters

    Parameters
    ----------
    filepath : str
        Filepath
    dataframe : pandas.DataFrame
        Dataframe to export into excel

    Examples
    --------
    >>> values = ["https://www.shawnngtq.com"]
    >>> cols = ["url"]
    >>> df = pd.DataFrame(values, columns=cols)
    >>> to_excel_keep_url("test_pandas_folder/tmp.xlsx", df)
    Excel exported ...

    Returns
    -------
    None
    """
    import pandas as pd

    if not isinstance(filepath, str):
        raise TypeError("Argument 1 must be a non-empty string ...")

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument 2 must be a dataframe ...")

    writer = pd.ExcelWriter(filepath, engine="xlsxwriter", options={"strings_to_urls": False})
    dataframe.to_excel(writer, index=False)
    writer.close()
    print("Excel exported ...")


def useless_columns(dataframe):
    """
    Return tuple of empty columns and single columns from given dataframe

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Base dataframe

    Returns
    -------
    empty_columns : list
        List of empty columns
    single_columns : list
        List of single value columns

    (empty_columns, single_columns)
    """
    import pandas as pd

    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError("Argument must be a dataframe ...")

    empty_columns = dataframe.nunique().where(lambda x: x == 0).dropna().index.values.tolist()
    single_columns = dataframe.nunique().where(lambda x: x == 1).dropna().index.values.tolist()
    print(f"Empty columns: {empty_columns}")
    print(f"Single value columns: {single_columns}")
    return empty_columns, single_columns


if __name__ == "__main__":
    import doctest
    import pandas as pd
    import subprocess

    subprocess.run("mkdir -p test_pandas_folder", shell=True, executable="/bin/bash")
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    subprocess.run("rm -rf test_pandas_folder", shell=True, executable="/bin/bash")



def haversine_np(longitude1, latitude1, longitude2, latitude2):
    """
    Return the great circle distance between two points on the earth (specified in decimal degrees). All args must be of equal length. Reference from https://stackoverflow.com/a/4913653, https://stackoverflow.com/a/29546836

    Parameters
    ----------
    longitude1: float
        Base longitude
    latitude1: float
        Base latitude
    longitude2: float
        Longitude to compare
    latitude2: float
        Latitude to compare

    Returns
    -------
    km : float
    """
    import numpy as np

    lon1, lat1, lon2, lat2 = map(np.radians, [longitude1, latitude1, longitude2, latitude2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2

    c = 2 * np.arcsin(np.sqrt(a))
    km = 6371 * c
    return km


if __name__ == "__main__":
    import doctest

    doctest.testmod()



def config_spark_local(autoset=True):
    """
    Automatically configure Spark local or provide recommendation if not autoset. Reference https://towardsdatascience.com/basics-of-apache-spark-configuration-settings-ca4faff40d45

    Parameters
    ----------
    autoset : bool
        whether to automatically configure spark

    Examples
    --------
    >>> config_spark_local()
    Here is the  current computer specs ...
    executor_per_node: 1
    spark_executor_instances: 1
    total_executor_memory: 30
    spark_executor_memory: 27
    memory_overhead: 3
    spark_default_parallelism: 10
    spark.sql.execution.arrow.pyspark.enabled recommended by Koalas ...
    spark auto-configured ...
    config_spark_local exited ...
    >>> config_spark_local(autoset=False)
    Here is the  current computer specs ...
    executor_per_node: 1
    spark_executor_instances: 1
    total_executor_memory: 30
    spark_executor_memory: 27
    memory_overhead: 3
    spark_default_parallelism: 10
    Here is the recommended command to execute:
    <BLANKLINE>
            spark = pyspark.sql.SparkSession.builder.master("local")             .config("spark.executor.cores", "5")             .config("spark.driver.cores", "5")             .config("spark.executor.instances", "1")             .config("spark.executor.memory", "27g")             .config("spark.driver.memory", "27g")             .config("spark.executor.memoryOverhead", "3g")             .config("spark.default.parallelism", "10")             .config("spark.sql.shuffle.partitions", "10")             .getOrCreate()
    <BLANKLINE>
    config_spark_local exited ...

    Returns
    -------
    None
    """
    import multiprocessing
    import math
    import os
    import pyspark

    def round_down_or_one(x):
        if math.floor(x) == 0:
            return 1
        else:
            return math.floor(x)

    print("Here is the  current computer specs ...")
    vcore_per_node = multiprocessing.cpu_count()
    spark_executor_cores = 5
    number_of_nodes = 1
    total_ram_per_node_gb = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES') / (1024. ** 3)

    executor_per_node = round_down_or_one((vcore_per_node - 1) / spark_executor_cores)
    spark_executor_instances = round_down_or_one((executor_per_node * number_of_nodes) - 1)
    total_executor_memory = round_down_or_one((total_ram_per_node_gb - 1) / executor_per_node)
    spark_executor_memory = round_down_or_one(total_executor_memory * 0.9)
    memory_overhead = round_down_or_one(total_executor_memory * 0.1)
    spark_default_parallelism = round_down_or_one(spark_executor_instances * spark_executor_cores * 2)
    print(f"executor_per_node: {executor_per_node}")
    print(f"spark_executor_instances: {spark_executor_instances}")
    print(f"total_executor_memory: {total_executor_memory}")
    print(f"spark_executor_memory: {spark_executor_memory}")
    print(f"memory_overhead: {memory_overhead}")
    print(f"spark_default_parallelism: {spark_default_parallelism}")

    if autoset:
        global spark
        spark = pyspark.sql.SparkSession.builder.master("local") \
            .config("spark.executor.cores", str(spark_executor_cores)) \
            .config("spark.driver.cores", str(spark_executor_cores)) \
            .config("spark.executor.instances", str(spark_executor_instances)) \
            .config("spark.executor.memory", f"{spark_executor_memory}g") \
            .config("spark.driver.memory", f"{spark_executor_memory}g") \
            .config("spark.executor.memoryOverhead", f"{memory_overhead}g") \
            .config("spark.default.parallelism", str(spark_default_parallelism)) \
            .config("spark.sql.shuffle.partitions", str(spark_default_parallelism)) \
            .getOrCreate()

        print("spark.sql.execution.arrow.pyspark.enabled recommended by Koalas ...")
        spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", True)
        spark.conf.get("spark.sql.execution.arrow.pyspark.enabled")
        print("spark auto-configured ...")
    else:
        print("Here is the recommended command to execute:")
        text = f"""
        spark = pyspark.sql.SparkSession.builder.master("local") \
            .config("spark.executor.cores", "{spark_executor_cores}") \
            .config("spark.driver.cores", "{spark_executor_cores}") \
            .config("spark.executor.instances", "{spark_executor_instances}") \
            .config("spark.executor.memory", "{spark_executor_memory}g") \
            .config("spark.driver.memory", "{spark_executor_memory}g") \
            .config("spark.executor.memoryOverhead", "{memory_overhead}g") \
            .config("spark.default.parallelism", "{spark_default_parallelism}") \
            .config("spark.sql.shuffle.partitions", "{spark_default_parallelism}") \
            .getOrCreate()
        """
        print(text)
    print("config_spark_local exited ...")


def add_dummy_columns(dataframe, columns, value):
    """
    Return dataframe with additional given dummy column(s) and value

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe
    columns : list
        List of column(s)
    value : str
        Default value of the new column(s)

    Returns
    -------
    df : pyspark.sql.dataframe.DataFrame
    """
    import pyspark
    import pyspark.sql.functions as F

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    if not isinstance(columns, list):
        raise TypeError("Argument must be a list ...")
    if not isinstance(value, str):
        raise TypeError("Argument must be a str ...")

    df = dataframe
    dummy_columns = set(columns) - set(dataframe.columns)
    for column in dummy_columns:
        df = df.withColumn(column, F.lit(value))
    return df


def column_into_list(dataframe, column):
    """
    Return list from given dataframe's column, with possible duplicates

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe
    column : str
        Column in dataframe

    Returns
    -------
    list_ : []
    """
    import pyspark

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    if not isinstance(column, str):
        raise TypeError("Argument must be a str ...")

    if column in dataframe.columns:
        list_ = dataframe.select(column).toPandas()[column].values.tolist()
        return list_


def column_into_set(dataframe, column):
    """
    Return normal set from given dataframe's column

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe
    column : str
        Column in dataframe

    Returns
    -------
    set_ : {}
    """
    import pyspark

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    if not isinstance(column, str):
        raise TypeError("Argument must be a str ...")

    set_ = set(column_into_list(dataframe, column))
    return set_


def columns_prefix(dataframe, prefix):
    """
    Return dataframe with renamed columns with given prefix

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe
    prefix : str
        Prefix

    Returns
    -------
    df : pyspark.sql.dataframe.DataFrame
    """
    import pyspark

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    if not isinstance(prefix, str):
        raise TypeError("Argument must be a str ...")

    df = dataframe
    for column in dataframe.columns:
        if not column.startswith(prefix):
            df = df.withColumnRenamed(column, prefix + column)
    return df


def columns_statistics(dataframe, n=10):
    """
    Display Spark dataframe columns' statistics and return tuple of empty columns and single columns from given dataframe

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe
    n : int / float
        Top n rows

    Returns
    -------
    empty_columns : list
        List of empty columns
    single_columns : list
        List of single value columns

    (empty_columns, single_columns)
    """
    import pyspark

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")

    describe(dataframe)
    empty_columns, single_columns = [], []

    for column in dataframe.columns:
        df = group_count(dataframe=dataframe, columns=column, n=n)
        print(column)
        df.show(n=n)

        if df.count() == 1:
            single_columns.append(column)
            print(f"!!!!! {column} is a candidate to drop !!!!!\n\n")

            if not df.first()[0] or df.first()[0].casefold() == "none" or df.first()[0].casefold():
                empty_columns.append(column)

    print(f"There are {len(single_columns)} of single value columns, they are: {single_columns}")
    print(f"There are {len(empty_columns)} of null value columns, they are: {empty_columns}")
    return empty_columns, single_columns


def describe(dataframe):
    """
    Display dataframe information similar to Pandas dataframe describe()

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe

    Returns
    -------
    None
    """
    import pyspark

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    print(f"The dataframe: {type(dataframe)}")
    print(f"Number of columns: {len(dataframe.columns)}")
    print(f"Number of rows: {dataframe.count()}")
    dataframe.printSchema()


def group_count(dataframe, columns, n=10):
    """
    Returns a dataframe group by column(s), sort in descending order, calculate count and percent

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        Spark dataframe
    columns : str / list
        List of column(s) to groupby
    n : int / float
        Top n rows

    Examples
    --------
    >>> data = {'id': [1, 1, 1, 2, 2, 3], 'value': [5, 2, 5, 2135, 124390, 213]}
    >>> df = spark.createDataFrame(pd.DataFrame(data))
    >>> group_count(df, ["id"]).show()
    +---+-----+-------+
    | id|count|percent|
    +---+-----+-------+
    |  1|    3|   50.0|
    |  2|    2| 33.333|
    |  3|    1| 16.667|
    +---+-----+-------+
    <BLANKLINE>
    >>> group_count(df, ["id", "value"]).show()
    +---+------+-----+-------+
    | id| value|count|percent|
    +---+------+-----+-------+
    |  1|     5|    2| 33.333|
    |  3|   213|    1| 16.667|
    |  2|  2135|    1| 16.667|
    |  1|     2|    1| 16.667|
    |  2|124390|    1| 16.667|
    +---+------+-----+-------+
    <BLANKLINE>

    Returns
    -------
    df : pyspark.sql.dataframe.DataFrame
    """
    import pyspark
    import pyspark.sql.functions as F

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    if not isinstance(columns, list):
        raise TypeError("Argument must be a list ...")

    df = dataframe.groupBy(columns).count().orderBy("count", ascending=False)
    row_count = dataframe.count()
    df = df.withColumn("percent", F.round(F.udf(lambda x: x * 100 / row_count)("count"), 3))
    if n != float("inf"):
        df = df.limit(n)
    return df


def rename(dataframe, columns):
    """
    Return dataframe with new renamed column(s)

    Parameters
    ----------
    dataframe : pyspark.sql.dataframe.DataFrame
        One Spark dataframe
    columns : dict
        A dictionary {oldName: newName} of columns to rename

    Examples
    --------
    >>> l = [('Alice', 1)]
    >>> df = spark.createDataFrame(l)
    >>> df.show()
    +-----+---+
    |   _1| _2|
    +-----+---+
    |Alice|  1|
    +-----+---+
    <BLANKLINE>
    >>> df2 = rename(df, {"_1": "name", "_2": "id"})
    >>> df2.show()
    +-----+---+
    | name| id|
    +-----+---+
    |Alice|  1|
    +-----+---+
    <BLANKLINE>

    Returns
    -------
    df : pyspark.sql.dataframe.DataFrame
    """
    import pyspark
    import pyspark.sql.functions as F

    if not isinstance(dataframe, pyspark.sql.dataframe.DataFrame):
        raise TypeError("Argument must be a Pyspark dataframe ...")
    if not isinstance(columns, dict):
        raise TypeError("Argument must be a dict ...")

    df = dataframe.select([F.col(c).alias(columns.get(c, c)) for c in dataframe.columns])
    return df


if __name__ == "__main__":
    import doctest
    import pandas as pd
    import pyspark

    spark = pyspark.sql.SparkSession.builder.master("local").getOrCreate()

    doctest.testmod()



def psy_pg_connect(host, database, username, password):
    """
    Return connection and cursor of psycopg2 postgres server connection

    Parameters
    ----------
    host: str
        Hostname
    database: str
        Database name
    username: str
        Username
    password: str
        Password

    Returns
    -------
    tuple
        (connection, cursor)
    """
    import psycopg2

    connection_string = f"host={host} port=5432 dbname={database} user={username} password={password}"
    connection = psycopg2.connect(connection_string)
    print("PostgreSQL database connected ...")

    cursor = connection.cursor()
    print("Cursor object created ...")
    return connection, cursor


def query_to_pandas(sql_query, connection):
    """
    Return Pandas dataframe from result of given SQL query

    Parameters
    ----------
    sql_query: str
        SQL query
    connection: str
        Connection return by psycopg2

    Examples
    --------
    >>> sql_query = '''SELECT * FROM users;'''

    Returns
    -------
    pandas.core.frame.DataFrame
    """
    import pandas as pd

    return pd.io.sql.read_sql_query(sql_query, connection)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



def init_django(project_name):
    """
    Setup Django in Jupyter. Reference from https://gist.github.com/codingforentrepreneurs/76e570d759f83d690bf36a8a8fa4cfbe

    Parameters
    ----------
    project_name: str
        Django project name

    Returns
    -------
    None
    """
    import os
    import sys
    import django

    present_working_dir = os.getenv("PWD")
    os.chdir(present_working_dir)

    project_name = project_name or os.environ.get("DJANGO_PROJECT") or None
    if not project_name:
        raise Exception("""
        Set an environment variable: `DJANGO_PROJECT=your_project_name` or call `init_django(your_project_name)`
        """)
    
    sys.path.insert(0, present_working_dir)
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"{project_name}.settings")
    os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
    django.setup()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from data_manipulation import (
    base, beautifulsoup_, django_, geopandas_, pandas_, psycopg2_, pyspark_,
)

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions



def build_soup(url):
    """
    Return Beautifulsoup object from given url

    Parameters
    ----------
    url: str
        URL

    Examples
    --------
    >>> a = build_soup("https://google.com")
    >>> type(a)
    <class 'bs4.BeautifulSoup'>

    Returns
    -------
    BS4 or None
    """
    import requests
    from bs4 import BeautifulSoup

    response = requests.get(url)
    if response.status_code == 200:
        soup = BeautifulSoup(response.text, "lxml")
        return soup
    else:
        print(f"response status code: {response.status_code}")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from setuptools import find_packages, setup

import versioneer

LONG_DESCRIPTION = """
Data Manipulation is a Python package providing powerful utility functions. 
It contains many subpackages with utility functions built for popular packages such as Pandas, PySpark and many more.
"""

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Environment :: Console",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Cython",
    "Topic :: Scientific/Engineering",
]

DEPENDENCIES = [
    "beautifulsoup4", "lxml",
    "django",
    "numpy",
    "pandas",
    "IPython", "requests",
    "psycopg2",
    "pyspark",
]

setup(
    author="Shawn Ng",
    author_email="shawn.coding.acc@gmail.com",
    classifiers=CLASSIFIERS,
    cmdclass=versioneer.get_cmdclass(),
    description="Powerful data manipulation",
    install_requires=DEPENDENCIES,
    license="BSD-3",
    long_description=LONG_DESCRIPTION,
    name="data_manipulation",
    packages=find_packages(include=["data_manipulation", "data_manipulation.*"]),
    platforms="any",
    test_suite="nose.collector",
    tests_require=["nose"],
    url="https://github.com/shawnngtq/data-manipulation",
    version=versioneer.get_version(),
)



def clean_string(string, remove_parenthesis=False, remove_brackets=False):
    """
    Return given string that is strip, uppercase without multiple whitespaces. Optionally, remove parenthesis and brackets. Note that "\t\n\s" will be removed

    Parameters
    ----------
    string : str
        String to clean
    remove_parenthesis : bool, optional
        To remove parenthesis
    remove_brackets : bool, optional
        To remove brackets

    Examples
    --------
    >>> clean_string(" sHawn  tesT ")
    'SHAWN TEST'
    >>> clean_string("shawn ( te  st )")
    'SHAWN ( TE ST )'
    >>> clean_string("shawn ( te  st )", remove_parenthesis=True)
    'SHAWN'
    >>> clean_string("shawn [ te  st ]", remove_brackets=True)
    'SHAWN'

    Returns
    -------
    string : str
    """
    import re

    if not isinstance(string, str):
        raise TypeError("Argument 1 must be str ...")

    if remove_parenthesis:
        string = re.sub(r"\(.*\)", "", string)
    if remove_brackets:
        string = re.sub(r"\[.*\]", "", string)

    string = string.strip().upper()
    string = " ".join(string.split())
    return string


def get_none_variation():
    """
    Get List of none variation

    Examples
    --------
    >>> get_none_variation()
    [None, 'none', 'None', 'NONE', 'null', 'Null', 'NULL', 'na', 'Na', 'nA', 'NA', 'N.A', 'N.A.', 'nil', 'Nil', 'NIL']

    Returns
    -------
    variations : list
    """
    variations = [
        None, "none", "None", "NONE",
        "null", "Null", "NULL",
        "na", "Na", "nA", "NA", "N.A", "N.A.",
        "nil", "Nil", "NIL",
    ]
    return variations


def get_path_files(path, keywords):
    """
    Return sorted list of files from given path and keywords

    Parameters
    ----------
    path : str
        Path
    keywords : list
        List of keywords

    Examples
    --------
    >>> get_path_files("test_base_folder", ["py"])
    ['test1.py', 'test2.py', 'test3.py', 'test4.py', 'test5.py']

    Returns
    -------
    list_ : list
    """
    import os

    if not isinstance(path, str):
        raise TypeError("Argument 1 must be str ...")
    if not isinstance(keywords, list):
        raise TypeError("Argument 2 must be list ...")

    list_ = []
    for file in sorted(os.listdir(path)):
        if any(word in file for word in keywords):
            list_.append(file)
    return list_


def list_tuple_without_none(list_tuple):
    """
    Return the given list / tuple without None variation

    Parameters
    ----------
    list_tuple : list or tuple
        [anyvalue, anytype, anylength]

    Examples
    --------
    >>> list_tuple_without_none(["a", "none"])
    ['a']
    >>> list_tuple_without_none(("a", "none"))
    ('a',)
    >>> list_tuple_without_none(get_none_variation())
    []
    >>> list_tuple_without_none(["a", "none", ""])
    ['a']
    >>> list_tuple_without_none(("a", "none", ""))
    ('a',)

    Returns
    -------
    lt : List or Tuple
    """
    none_variations = get_none_variation()
    if isinstance(list_tuple, list):
        lt = [item for item in list_tuple if item and item not in none_variations]
    elif isinstance(list_tuple, tuple):
        lt = tuple(item for item in list_tuple if item and item not in none_variations)
    else:
        raise TypeError("Argument must be list or tuple ...")
    return lt


def remove_path_file(path, keyword, n=2):
    """
    Remove (n-2) oldest files from given path and keyword

    Parameters
    ----------
    path: str
        Directory path
    keyword: str
        Keyword
    n: int, optional
        Keep latest n files

    Examples
    --------
    >>> get_path_files("test_base_folder", ["py"])
    ['test1.py', 'test2.py', 'test3.py', 'test4.py', 'test5.py']
    >>> remove_path_file("test_base_folder", ".py")
    test_base_folder/test1.py deleted ...
    test_base_folder/test2.py deleted ...
    test_base_folder/test3.py deleted ...
    >>> get_path_files("test_base_folder", ["py"])
    ['test4.py', 'test5.py']

    Returns
    -------
    None
    """
    import os

    if not isinstance(path, str):
        raise TypeError("Argument 1 must be str ...")
    if not isinstance(keyword, str):
        raise TypeError("Argument 2 must be str ...")
    if not isinstance(n, int):
        raise TypeError("Argument 3 must be int ...")

    to_delete = get_path_files(path=path, keywords=[keyword])[:-n]
    for file in to_delete:
        os.remove(f"{path}/{file}")
        print(f"{path}/{file} deleted ...")


def string_boolean_to_int(boolean_str_rep):
    """
    Return integer from given string boolean. 1 instead of "true"/"True"/"1". Reference from https://docs.python.org/3/distutils/apiref.html#distutils.util.strtobool

    Parameters
    ----------
    boolean_str_rep : str
        "True"

    Notes
    -----
    Do not use eval() as it's unsafe

    Examples
    --------
    >>> string_boolean_to_int("true")
    1
    >>> string_boolean_to_int("True")
    1
    >>> string_boolean_to_int("1")
    1

    Returns
    -------
    int_ : int
    """
    from distutils.util import strtobool

    if not isinstance(boolean_str_rep, str):
        raise TypeError("Argument must be str ...")
    int_ = strtobool(string_str_to_str(boolean_str_rep))
    return int_


def string_dlt_to_dlt(dlt_str_rep):
    """
    Return dictionary/list/tuple from given string representation of dictionary/list/tuple

    Parameters
    ----------
    dlt_str_rep : str
        '[]'

    Examples
    --------
    >>> string_dlt_to_dlt("")
    Traceback (most recent call last):
      ...
    SyntaxError: unexpected EOF while parsing
    >>> string_dlt_to_dlt("[1, 2, 3]")
    [1, 2, 3]
    >>> string_dlt_to_dlt("[]")
    []
    >>> string_dlt_to_dlt("['1', '2', '3']")
    ['1', '2', '3']
    >>> string_dlt_to_dlt("{'a': 1, 'b': 2}")
    {'a': 1, 'b': 2}
    >>> string_dlt_to_dlt("{'a': '1', 'b': '2'}")
    {'a': '1', 'b': '2'}
    >>> string_dlt_to_dlt("('1', '2', '3')")
    ('1', '2', '3')

    Returns
    -------
    dlt : dict or list or tuple
    """
    from ast import literal_eval

    if not isinstance(dlt_str_rep, str):
        raise TypeError("Argument must be str ...")
    dlt = literal_eval(dlt_str_rep)
    return dlt


def string_str_to_str(string_str_rep):
    """
    Return string from given string representation of string

    Parameters
    ----------
    string_str_rep : str
        "'str'"

    Examples
    --------
    >>> string_str_to_str("")
    ''
    >>> string_str_to_str('test')
    'test'
    >>> string_str_to_str("'test'")
    'test'
    >>> string_str_to_str('"test"')
    'test'

    Returns
    -------
    str_ : str
    """
    if not isinstance(string_str_rep, str):
        raise TypeError("Argument must be str ...")
    str_ = string_str_rep.strip("'\"")
    return str_


if __name__ == "__main__":
    import doctest
    import subprocess

    subprocess.run("mkdir -p test_base_folder", shell=True, executable="/bin/bash")
    subprocess.run("touch test_base_folder/test{1..5}.py", shell=True, executable="/bin/bash")
    doctest.testmod()
    subprocess.run("rm -rf test_base_folder", shell=True, executable="/bin/bash")
